import json
from .coco import *
import copy

def coco_merge(*coco_dicts: dict, info: COCOInfo = COCOInfo()) -> dict:
    """
    Merge COCO annotations.

    adapted from pyodi https://github.com/Gradiant/pyodi under MPL 2.0

    Parameters
    ----------
    *coco_files: dicts
        dicts from coco files as e.g. read by json.load
    
    Returns
    -------
    coco_data: dict
        coco data for merged output as dict
    """

    categories = []
    images = []
    annotations = []
    licenses = []

    license_map = {}
    category_map = {}
    image_map = {}

    for data in coco_dicts:
        
        cat_id_map = {}
        for new_cat in data["categories"]:
            label = new_cat["name"]
            if label in category_map:
                cat_id_map[new_cat["id"]] = category_map[label]
            else:
                new_id = len(categories) + 1
                cat_id_map[new_cat["id"]] = new_id
                new_cat["id"] = new_id
                categories.append(new_cat)
                category_map[label] = new_id

        license_id_map = {}
        for new_license in data["licenses"]:
            license_name = new_license["name"]
            if license_name in license_map:
                license_id_map[new_license["id"]] = license_map[license_name]
            else:
                new_id = len(licenses) + 1
                license_id_map[new_license["id"]] = new_id
                new_license["id"] = new_id
                licenses.append(new_license)
                license_map[license_name] = new_id

        image_id_map = {}
        for new_image in data["images"]:
            file_name = new_image["file_name"]
            if file_name in image_map:
                image_id_map[new_image["id"]] = image_map[file_name]
            else:
                new_id = len(images) + 1
                image_id_map[new_image["id"]] = new_id
                new_image["id"] = new_id
                if "license_id" in new_image:
                    new_image["license_id"] = license_id_map[new_image["license_id"]]
                images.append(new_image)
                image_map[file_name] = new_id

        for new_annotation in data["annotations"]:
            new_id = len(annotations) + 1
            new_annotation["id"] = new_id
            new_annotation["category_id"] = cat_id_map[new_annotation["category_id"]]
            new_annotation["image_id"] = image_id_map[new_annotation["image_id"]]
            annotations.append(new_annotation)

    return {
        "annotations": annotations,
        "images": images,
        "info": info.to_dict(),
        "categories": categories,
        "licenses": licenses
    }

def coco_remap_categories(coco_dict: dict, name_mapping: dict) -> dict:
    """
    Rename categories based on a mapping. Does NOT merge or re-index.
    Resulting dataset may contain multiple categories with the same name.
    
    Parameters
    ----------
    coco_dict: dict
        Source dataset
    name_mapping: dict
        Old Name -> New Name. Unlisted names are kept as-is.

    Returns
    -------
    dict
        The original dataset, with category names remapped

    Examples
    --------
    >>> data = {
    ...     'categories': [
    ...         {'id': 1, 'name': 'taxi'},
    ...         {'id': 2, 'name': 'bus'},
    ...         {'id': 3, 'name': 'dog'}
    ...     ]
    ... }
    >>> mapping = {'taxi': 'vehicle', 'bus': 'vehicle'}
    >>> _ = coco_remap_categories(data, mapping)
    >>> import pprint; pprint.pprint(data['categories'])
    [{'id': 1, 'name': 'vehicle'},
     {'id': 2, 'name': 'vehicle'},
     {'id': 3, 'name': 'dog'}]
    """
    for cat in coco_dict.get("categories", []):
        if cat["name"] in name_mapping:
            cat["name"] = name_mapping[cat["name"]]

    return coco_dict

def coco_collapse_categories(coco_dict: dict) -> dict:
    """
    Merge categories that share the same name into a single ID.
    Updates annotations to point to the unified ID.

    Parameters
    ----------
    coco_dict: dict
        Source dataset

    Returns
    -------
    dict
        The original dataset, with duplicate categories collapsed

    Examples
    --------
    >>> data = {
    ...     'categories': [
    ...         {'id': 10, 'name': 'vehicle'},
    ...         {'id': 20, 'name': 'vehicle'}
    ...     ],
    ...     'annotations': [
    ...         {'id': 1, 'category_id': 10},
    ...         {'id': 2, 'category_id': 20}
    ...     ]
    ... }
    >>> _ = coco_collapse_categories(data)
    >>> # Notice ID 20 remains because it was last in the list
    >>> data['categories']
    [{'id': 20, 'name': 'vehicle'}]
    >>> # Both annotations now point to 20
    >>> [ann['category_id'] for ann in data['annotations']]
    [20, 20]
    """

    name_to_target_id = {cat["name"]: cat["id"] for cat in coco_dict["categories"]}
    all_target_ids = set(name_to_target_id.values())
    id_remap = {cat["id"]: name_to_target_id[cat["name"]] for cat in coco_dict["categories"]}
    coco_dict["categories"] = [cat for cat in coco_dict["categories"] if cat["id"] in all_target_ids]
    for ann in coco_dict["annotations"]:
        ann["category_id"] = id_remap[ann["category_id"]]

    return coco_dict

def coco_reindex_categories(coco_dict: dict) -> dict:
    """
    Re-index category IDs to be contiguous integers starting from 1.
    
    Parameters
    ----------
    coco_dict: dict
        Source dataset

    Returns
    -------
    dict
        The original dataset, with categories reindexed

    Examples
    --------
    >>> data = {
    ...     'categories': [{'id': 20, 'name': 'vehicle'}],
    ...     'annotations': [{'id': 1, 'category_id': 20}]
    ... }
    >>> _ = coco_reindex_categories(data)
    >>> data['categories']
    [{'id': 1, 'name': 'vehicle'}]
    >>> data['annotations']
    [{'id': 1, 'category_id': 1}]
    """
    id_remap = {cat["id"]: i+1 for i, cat in enumerate(coco_dict["categories"])}
    for cat in coco_dict["categories"]:
        cat["id"] = id_remap[cat["id"]]
    for ann in coco_dict["annotations"]:
        ann["category_id"] = id_remap[ann["category_id"]]

    return coco_dict



import pycocotools.coco
import json
from collections.abc import Iterable
import csv

COCO2CSV_ANNOTATION_HEADER = ['id', 'image_id', 'category_id', 'segmentation', 'area', 'bbox', 'iscrowd', 'keypoints', 'num_keypoints', 'caption', 'file_name', 'segments_info']
COCO2CSV_IMAGE_HEADER = ['width', 'height', 'file_name', 'flickr_url', 'coco_url', 'date_captured']
COCO2CSV_CATEGORY_HEADER = ['name', 'supercategory', 'keypoints', 'skeleton', 'isthing', 'color']

def csv_safe(val):
    """
    Convert a value to a CSV-safe representation.

    Lists and dictionaries are converted to JSON strings. 
    All other values are returned unchanged.

    Parameters
    ----------
    val : Any
        The value to be converted. Can be of any type.

    Returns
    -------
    Any
        If `val` is a list or dictionary, returns a JSON-encoded string.
        Otherwise, returns the original value.

    Examples
    --------
    >>> csv_safe([1, 2, 3])
    '[1, 2, 3]'
    >>> csv_safe({'x': 1})
    '{"x": 1}'
    >>> csv_safe(42)
    42
    """
    if isinstance(val, (list, dict)):
        return json.dumps(val)
    return val

def read_header(ann: dict, img: dict, cat: dict) -> list:
    """
    Determine which columns exist in annotation, image, and category dictionaries.

    Filters the predefined COCO CSV headers to those keys actually present 
    in the provided dictionaries.

    Parameters
    ----------
    ann : dict
        A COCO annotation dictionary.
    img : dict
        A COCO image dictionary corresponding to `ann['image_id']`.
    cat : dict
        A COCO category dictionary corresponding to `ann['category_id']`.

    Returns
    -------
    tuple of list
        Three lists of column names:
        - Annotation columns present in `ann`
        - Image columns present in `img`
        - Category columns present in `cat`

    Examples
    --------
    >>> ann = {'id': 1, 'image_id': 5, 'bbox': [0,0,10,10]}
    >>> img = {'width': 640, 'height': 480, 'file_name': 'img.jpg'}
    >>> cat = {'name': 'person', 'supercategory': 'human'}
    >>> read_header(ann, img, cat)
    (['id', 'image_id', 'bbox'], ['width', 'height', 'file_name'], ['name', 'supercategory'])
    """
    return (
        [col for col in COCO2CSV_ANNOTATION_HEADER if col in ann],
        [col for col in COCO2CSV_IMAGE_HEADER      if col in img],
        [col for col in COCO2CSV_CATEGORY_HEADER   if col in cat]
    )

def flatten_coco(coco: pycocotools.coco.COCO) -> Iterable[list]:
    """
    Generate CSV rows from a COCO-format JSON annotation file.

    Each row combines fields from the annotation, the associated image, 
    and the associated category. Lists and dictionaries are JSON-encoded 
    for CSV compatibility. The first row yielded is a header row.

    Parameters
    ----------
    filename : pycocotools.coco.COCO
        The loaded COCO annotations.

    Yields
    ------
    list
        A row of CSV data as a list of values. The first row is the header.

    Examples
    --------
    >>> coco = pycocotools.coco.COCO('tests/example_coco.json')
    loading annotations into memory...
    Done (t=0.00s)
    creating index...
    index created!
    >>> flat_coco = flatten_coco(coco)
    >>> import itertools
    >>> for row in itertools.islice(flat_coco, 4):
    ...     print(row)
    ['id', 'image_id', 'category_id', 'area', 'bbox', 'iscrowd', 'width', 'height', 'file_name', 'name']
    [1, 1, 60, 102225.5616, '[117.84031999999999, 235.14, 425.93984, 240.0]', 0, 640, 480, '000000000612.jpg', 'bed']
    [2, 1, 78, 6234.393046425599, '[320.4, 271.38984000000005, 71.07968, 87.70992]', 0, 640, 480, '000000000612.jpg', 'teddy bear']
    [3, 1, 78, 5028.19008, '[260.31968, 281.72016, 62.0, 81.09984]', 0, 640, 480, '000000000612.jpg', 'teddy bear']
    """
    headers = None
    for ann in coco.anns.values():
        img = coco.imgs[ann['image_id']]
        cat = coco.cats[ann['category_id']]
        if headers is None:
            headers = read_header(ann, img, cat)
            yield [*headers[0], *headers[1], *headers[2]]
        yield [
            csv_safe(obj[col]) 
            for obj, head in zip([ann, img, cat], headers) 
            for col in head
        ]

def coco2csv(coco_path: str, csv_path: str, **csv_writer_kwargs) -> None:
    """
    Convenience wrapper to write a coco json file directly to a csv file.
    Recommend that users use `flatten_coco` instead, and then manually adjust 
    the csv fields to align with their expected format before writing.

    Parameters
    ----------
    coco_path: str
    The path to the coco json file
  
    csv_path: str
    The path where the result csv should be written

    Returns
    -------
    returns None, but writes a file to disk at csv_path
    """
    coco = pycocotools.coco.COCO(coco_path)
    flat_coco = flatten_coco(coco)
    with open(csv_path, 'w') as f:
        writer = csv.writer(csv_path, **csv_writer_kwargs)
        writer.writerows(flat_coco)



import unittest
import utils
import pycocowriter.coco2csv

def load_tests(loader, tests, ignore):
    return utils.doctests(pycocowriter.coco2csv, tests)


if __name__ == '__main__':
    unittest.main()
